import logging
from typing import Union

from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class DeviceInfoTableTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:

        deviceId = tupleSelector.selector.get("deviceId")

        ormSession = self._ormSessionCreator()
        try:
            query = ormSession.query(
                DeviceInfoTable, GpsLocationTable
            ).outerjoin(GpsLocationTable)

            if deviceId is not None:
                query = query.filter(DeviceInfoTable.deviceId == deviceId)

            # Add the current location to each DeviceInfo row
            tuples = []
            for deviceInfoTableRow, gpsLocationTableRow in query.all():
                deviceInfoTableRow.currentLocation = gpsLocationTableRow
                tuples.append(deviceInfoTableRow)

            # Create the vortex message
            return (
                Payload(filt, tuples=tuples).makePayloadEnvelope().toVortexMsg()
            )

        finally:
            ormSession.close()



import logging
from pathlib import Path

from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.TupleAction import TupleActionABC
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device._private.server.controller.EnrollmentController import (
    EnrollmentController,
)
from peek_core_device._private.server.controller.GpsController import (
    GpsController,
)
from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.server.controller.DeviceStatusController import (
    DeviceStatusController,
)
from peek_core_device._private.server.controller.OfflineCacheController import (
    OfflineCacheController,
)
from peek_core_device._private.server.controller.UpdateController import (
    UpdateController,
)

logger = logging.getLogger(__name__)


class MainController(TupleActionProcessorDelegateABC):
    def __init__(
        self,
        dbSessionCreator,
        notifierController: NotifierController,
        offlineCacheController: OfflineCacheController,
        deviceUpdateFilePath: Path,
        tupleObservable: TupleDataObservableHandler,
    ):
        self._dbSessionCreator = dbSessionCreator
        self._notifierController = notifierController

        self._enrollmentController = EnrollmentController(
            dbSessionCreator, notifierController
        )

        self._onlineController = DeviceStatusController(
            dbSessionCreator, notifierController
        )

        self._updateController = UpdateController(
            dbSessionCreator, notifierController, deviceUpdateFilePath
        )

        self._gpsController = GpsController(
            dbSessionCreator=dbSessionCreator,
            tupleObservable=tupleObservable,
            notifierController=notifierController,
        )

        self._offlineCacheController = offlineCacheController

    @property
    def deviceUpdateController(self):
        return self._updateController

    def shutdown(self):
        self._enrollmentController.shutdown()
        self._onlineController.shutdown()
        self._updateController.shutdown()
        self._gpsController.shutdown()

    @inlineCallbacks
    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        result = yield self._enrollmentController.processTupleAction(
            tupleAction
        )
        if result is not None:
            return result

        result = yield self._onlineController.processTupleAction(tupleAction)
        if result is not None:
            return result

        result = yield self._updateController.processTupleAction(tupleAction)
        if result is not None:
            return result

        result = yield self._gpsController.processTupleAction(tupleAction)
        if result is not None:
            return result

        result = yield self._offlineCacheController.processTupleAction(
            tupleAction
        )
        if result is not None:
            return result

        raise NotImplementedError(tupleAction.tupleName())



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.PluginNames import deviceTuplePrefix
from peek_core_device._private.tuples.OfflineCacheStatusTuple import (
    OfflineCacheStatusTuple,
)


@addTupleType
class OfflineCacheStatusAction(TupleActionABC):
    """Offline Cache Status Action

    This tuple is sent from the client to the server periodically when
    an update cycle finishes

    """

    __tupleType__ = deviceTuplePrefix + "OfflineCacheStatusAction"

    deviceToken: str = TupleField()
    cacheStatusList: list[OfflineCacheStatusTuple] = TupleField([])



from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

from peek_core_device._private.PluginNames import deviceFilt
from peek_core_device._private.PluginNames import deviceObservableName
from peek_core_device._private.server.controller.OfflineCacheController import (
    OfflineCacheController,
)
from peek_core_device._private.server.tuple_providers.ClientSettingsTupleProvider import (
    ClientSettingsTupleProvider,
)
from peek_core_device._private.server.tuple_providers.DeviceCacheStatusTupleProvider import (
    DeviceCacheStatusTupleProvider,
)
from peek_core_device._private.server.tuple_providers.DeviceGpsLocationTupleProvider import (
    DeviceGpsLocationTupleProvider,
)
from peek_core_device._private.server.tuple_providers.DeviceInfoTableTupleProvider import (
    DeviceInfoTableTupleProvider,
)
from peek_core_device._private.server.tuple_providers.DeviceInfoTupleProvider import (
    DeviceInfoTupleProvider,
)
from peek_core_device._private.server.tuple_providers.DeviceUpdateTupleProvider import (
    DeviceUpdateTupleProvider,
)
from peek_core_device._private.server.tuple_providers.OfflineCacheSettingTupleProvider import (
    OfflineCacheSettingTupleProvider,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.DeviceUpdateTuple import (
    DeviceUpdateTuple,
)
from peek_core_device._private.tuples.ClientSettingsTuple import (
    ClientSettingsTuple,
)
from peek_core_device._private.tuples.DeviceCacheStatusTuple import (
    DeviceCacheStatusTuple,
)
from peek_core_device._private.tuples.OfflineCacheSettingTuple import (
    OfflineCacheSettingTuple,
)
from peek_core_device.tuples.DeviceGpsLocationTuple import (
    DeviceGpsLocationTuple,
)
from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple


def makeTupleDataObservableHandler(
    ormSessionCreator, offlineCacheController: OfflineCacheController
):
    """ " Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param ormSessionCreator: A function that returns a SQLAlchemy session when called
    :param offlineCacheController:

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=deviceObservableName, additionalFilt=deviceFilt
    )

    # Register TupleProviders here
    tupleObservable.addTupleProvider(
        DeviceUpdateTuple.tupleName(),
        DeviceUpdateTupleProvider(ormSessionCreator),
    )

    tupleObservable.addTupleProvider(
        DeviceInfoTuple.tupleName(),
        DeviceInfoTupleProvider(ormSessionCreator, offlineCacheController),
    )

    tupleObservable.addTupleProvider(
        DeviceCacheStatusTuple.tupleName(),
        DeviceCacheStatusTupleProvider(offlineCacheController),
    )

    tupleObservable.addTupleProvider(
        DeviceInfoTable.tupleName(),
        DeviceInfoTableTupleProvider(ormSessionCreator),
    )

    tupleObservable.addTupleProvider(
        ClientSettingsTuple.tupleName(),
        ClientSettingsTupleProvider(ormSessionCreator),
    )

    tupleObservable.addTupleProvider(
        DeviceGpsLocationTuple.tupleName(),
        DeviceGpsLocationTupleProvider(ormSessionCreator),
    )

    tupleObservable.addTupleProvider(
        OfflineCacheSettingTuple.tupleName(),
        OfflineCacheSettingTupleProvider(ormSessionCreator),
    )

    return tupleObservable



import logging
from typing import Union

from peek_core_device._private.server.controller.OfflineCacheController import (
    OfflineCacheController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device._private.tuples.DeviceCacheStatusTuple import (
    DeviceCacheStatusTuple,
)

logger = logging.getLogger(__name__)


class DeviceCacheStatusTupleProvider(TuplesProviderABC):
    def __init__(self, offlineCacheController: OfflineCacheController):
        self._offlineCacheController = offlineCacheController

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:

        deviceToken = tupleSelector.selector.get("deviceToken")

        tuple_ = DeviceCacheStatusTuple(
            statusList=self._offlineCacheController.lastCacheStatus(deviceToken)
        )
        # Create the vortex message
        return (
            Payload(filt, tuples=[tuple_]).makePayloadEnvelope().toVortexMsg()
        )



import logging
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device._private.storage.Setting import FIELD_ENROLLMENT_ENABLED
from peek_core_device._private.storage.Setting import OFFICE_ENROLLMENT_ENABLED
from peek_core_device._private.storage.Setting import globalSetting
from peek_core_device._private.tuples.ClientSettingsTuple import \
    ClientSettingsTuple

logger = logging.getLogger(__name__)


class ClientSettingsTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:

        ormSession = self._ormSessionCreator()
        try:
            clientSetting = ClientSettingsTuple()

            clientSetting.fieldEnrollmentEnabled = globalSetting(
                ormSession, FIELD_ENROLLMENT_ENABLED
            )

            clientSetting.officeEnrollmentEnabled = globalSetting(
                ormSession, OFFICE_ENROLLMENT_ENABLED
            )

            # Create the vortex message
            return (
                Payload(filt, tuples=[clientSetting])
                    .makePayloadEnvelope()
                    .toVortexMsg()
            )

        finally:
            ormSession.close()



import logging
from datetime import datetime
from typing import Optional

from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.Tuple import Tuple

from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple
from sqlalchemy import Column
from sqlalchemy.sql.sqltypes import Boolean
from sqlalchemy.sql.sqltypes import DateTime
from sqlalchemy.sql.sqltypes import Integer
from sqlalchemy.sql.sqltypes import String

from .DeclarativeBase import DeclarativeBase
from ..PluginNames import deviceTuplePrefix
from ...tuples.DeviceGpsLocationTuple import DeviceGpsLocationTuple

logger = logging.getLogger(__name__)


@addTupleType
class DeviceInfoTable(Tuple, DeclarativeBase):
    """DeviceInfoTable

    This table stores information about devices.

    """

    __tablename__ = "DeviceInfo"
    __tupleType__ = deviceTuplePrefix + "DeviceInfoTable"

    TYPE_FIELD_IOS = DeviceInfoTuple.TYPE_FIELD_IOS
    TYPE_FIELD_ANDROID = DeviceInfoTuple.TYPE_FIELD_ANDROID
    TYPE_FIELD_WEB = DeviceInfoTuple.TYPE_FIELD_WEB
    TYPE_OFFICE_WEB = DeviceInfoTuple.TYPE_OFFICE_WEB
    TYPE_DESKTOP_WINDOWS = DeviceInfoTuple.TYPE_DESKTOP_WINDOWS
    TYPE_DESKTOP_MACOS = DeviceInfoTuple.TYPE_DESKTOP_MACOS

    DEVICE_OFFLINE = DeviceInfoTuple.DEVICE_OFFLINE
    DEVICE_ONLINE = DeviceInfoTuple.DEVICE_ONLINE
    DEVICE_BACKGROUND = DeviceInfoTuple.DEVICE_BACKGROUND

    id = Column(Integer, primary_key=True)
    description = Column(String(100), nullable=False, unique=True)
    deviceId = Column(String(50), nullable=False, unique=True)
    deviceType = Column(String(20), nullable=False)
    deviceToken = Column(String(50), nullable=False, unique=True)
    appVersion = Column(String(15), nullable=False)
    updateVersion = Column(String(15))  # Null means it hasn't updated
    lastOnline = Column(DateTime(True))
    lastUpdateCheck = Column(DateTime(True))
    createdDate = Column(DateTime(True), nullable=False)
    deviceStatus = Column(Integer, nullable=False, server_default="0")
    isEnrolled = Column(Boolean, nullable=False, server_default="0")
    isOfflineCacheEnabled = Column(Boolean, nullable=False, server_default="0")
    currentLocation: DeviceGpsLocationTuple = TupleField()
    lastCacheUpdate: Optional[datetime] = TupleField()

    def toTuple(
        self,
        currentLocationTuple: DeviceGpsLocationTuple = None,
        lastCacheCheck: Optional[datetime] = None,
    ):
        return self.toTupleStatic(self, currentLocationTuple, lastCacheCheck)

    @staticmethod
    def toTupleStatic(
        table: "DeviceInfoTable",
        currentLocationTuple: Optional[DeviceGpsLocationTuple] = None,
        lastCacheCheck: Optional[datetime] = None,
    ):
        return DeviceInfoTuple(
            description=table.description,
            deviceId=table.deviceId,
            deviceType=table.deviceType,
            deviceToken=table.deviceToken,
            appVersion=table.appVersion,
            updateVersion=table.updateVersion,
            lastOnline=table.lastOnline,
            lastUpdateCheck=table.lastUpdateCheck,
            createdDate=table.createdDate,
            deviceStatus=table.deviceStatus,
            isEnrolled=table.isEnrolled,
            currentLocation=currentLocationTuple,
            lastCacheCheck=lastCacheCheck,
        )



import logging
from typing import List

from twisted.internet.task import LoopingCall
from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexFactory import VortexFactory
from vortex.handler.TupleDataActionClient import TupleDataActionClient
from vortex.handler.TupleDataObservableProxyHandler import (
    TupleDataObservableProxyHandler,
)

from peek_core_device._private.PluginNames import deviceActionProcessorName
from peek_core_device._private.PluginNames import deviceFilt
from peek_core_device._private.PluginNames import deviceObservableName
from peek_core_device._private.tuples.UpdateDeviceOnlineAction import (
    UpdateDeviceOnlineAction,
)
from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger
from peek_plugin_base.PeekVortexUtil import peekServerName

from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple

logger = logging.getLogger(__name__)


def makeDeviceTupleDataObservableProxy():
    return TupleDataObservableProxyHandler(
        observableName=deviceObservableName,
        proxyToVortexName=peekServerName,
        additionalFilt=deviceFilt,
    )


deviceOnlineFilt = dict(key="device.online")
deviceOnlineFilt.update(deviceFilt)


# The sender is located at:
# peek_core_device/plugin-module/_private/device-server.service.ts


class DeviceOnlineHandler:
    CHECK_TIME = 5.0

    def __init__(self):
        self._ep = PayloadEndpoint(deviceOnlineFilt, self._process)
        self._onlineDeviceIdsByUuid = {}

        self._loopingCall = LoopingCall(self._poll)
        self._loopingCall.start(self.CHECK_TIME, now=False)

        self._actionClient = TupleDataActionClient(
            tupleActionProcessorName=deviceActionProcessorName,
            destVortexName=peekServerName,
            additionalFilt=deviceFilt,
        )

    def shutdown(self):
        self._ep.shutdown()
        self._ep = None

        self._loopingCall.stop()

        # This is never going to send, the server needs to do it
        # self._sendOfflineForDeviceIds(self._onlineDeviceIdsByUuid.values())
        self._onlineDeviceIdsByUuid = {}

    @peekCatchErrbackWithLogger(logger)
    def _poll(self):
        try:
            onlineVortexUuids = set(VortexFactory.getRemoteVortexUuids())
            onlineDeviceVortexUuids = set(self._onlineDeviceIdsByUuid)

            devicesThatHaveGoneOffline = (
                onlineDeviceVortexUuids - onlineVortexUuids
            )

            deviceIds = []
            for vortexUuid in devicesThatHaveGoneOffline:
                deviceIds.append(self._onlineDeviceIdsByUuid.pop(vortexUuid))

            # Make sure the device hasn't come back online as a difference vortex UUID
            deviceIds = list(
                set(deviceIds) - set(self._onlineDeviceIdsByUuid.values())
            )

            self._sendOfflineForDeviceIds(deviceIds)

        except Exception as e:
            logger.exception(e)

    def _process(
        self, payloadEnvelope: PayloadEnvelope, vortexUuid: str, **kwargs
    ):
        deviceId = payloadEnvelope.filt["deviceId"]

        if vortexUuid in self._onlineDeviceIdsByUuid:
            return

        self._onlineDeviceIdsByUuid[vortexUuid] = deviceId

        action = UpdateDeviceOnlineAction()
        action.deviceId = deviceId
        action.deviceStatus = DeviceInfoTuple.DEVICE_ONLINE

        d = self._actionClient.pushAction(action)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def _sendOfflineForDeviceIds(self, deviceIds: List[str]):
        for deviceId in deviceIds:
            action = UpdateDeviceOnlineAction()
            action.deviceId = deviceId
            action.deviceStatus = DeviceInfoTuple.DEVICE_OFFLINE

            d = self._actionClient.pushAction(action)
            d.addErrback(vortexLogFailure, logger, consumeError=True)



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.PluginNames import deviceTuplePrefix


@addTupleType
class UpdateEnrollmentAction(TupleActionABC):
    """Authorise Enrolment Action

    This action authorises a device to enroll in this peek environment.

    """

    __tupleType__ = deviceTuplePrefix + "UpdateEnrollmentAction"

    #:  The device info id to authorise
    deviceInfoId: int = TupleField()
    unenroll: bool = TupleField(False)
    remove: bool = TupleField(False)



import logging
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable

logger = logging.getLogger(__name__)


class DeviceGpsLocationTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        deviceToken = tupleSelector.selector.get("deviceToken")

        ormSession = self._ormSessionCreator()
        try:
            query = ormSession.query(GpsLocationTable)

            if deviceToken:
                query = query.filter(
                    GpsLocationTable.deviceToken == deviceToken)

            tuples = [t.toTuple() for t in query.all()]

            return Payload(filt,
                tuples=tuples).makePayloadEnvelope().toVortexMsg()
        finally:
            ormSession.close()



import logging
from typing import List

from sqlalchemy.orm.exc import NoResultFound
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Tuple import Tuple
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.tuples.DeviceBackgroundStateTupleAction import (
    DeviceBackgroundStateTupleAction,
)
from peek_core_device._private.tuples.UpdateDeviceOnlineAction import (
    UpdateDeviceOnlineAction,
)
from peek_core_device.tuples.DeviceInfoTuple import DeviceInfoTuple

logger = logging.getLogger(__name__)


class DeviceStatusController:
    def __init__(self, dbSessionCreator, notifierController: NotifierController):
        self._dbSessionCreator = dbSessionCreator
        self._notifierController = notifierController

        self._setAllDevicesOfflineBlocking()

    def shutdown(self):
        pass

    def processTupleAction(self, tupleAction: TupleActionABC) -> List[Tuple]:
        if isinstance(tupleAction, UpdateDeviceOnlineAction):
            return self._processUpdateOnline(tupleAction)

        if isinstance(tupleAction, DeviceBackgroundStateTupleAction):
            return self._processUpdateBackgrounded(tupleAction)

    @deferToThreadWrapWithLogger(logger)
    def _processUpdateBackgrounded(
        self, action: DeviceBackgroundStateTupleAction
    ) -> List[Tuple]:
        """Process Device Backgrounded Update

        :rtype: Deferred
        """
        session = self._dbSessionCreator()

        try:
            deviceInfo = (
                session.query(DeviceInfoTable)
                .filter(DeviceInfoTable.deviceId == action.deviceId)
                .one()
            )

            deviceId = deviceInfo.deviceId
            deviceInfo.lastOnline = action.dateTime

            if deviceInfo.deviceStatus & DeviceInfoTuple.DEVICE_ONLINE:
                deviceInfo.deviceStatus = DeviceInfoTuple.DEVICE_ONLINE
            else:
                deviceInfo.deviceStatus = DeviceInfoTuple.DEVICE_OFFLINE

            if action.deviceBackgrounded:
                deviceInfo.deviceStatus |= DeviceInfoTuple.DEVICE_BACKGROUND

            session.commit()

            self._notifierController.notifyDeviceInfo(deviceId=deviceId)
            self._notifierController.notifyDeviceOnline(
                deviceInfo.deviceId, deviceInfo.deviceToken, deviceInfo.deviceStatus
            )

            return []

        except NoResultFound:
            return []

        finally:
            session.close()

    @deferToThreadWrapWithLogger(logger)
    def _processUpdateOnline(self, action: UpdateDeviceOnlineAction) -> List[Tuple]:
        """Process Online Status Update

        :rtype: Deferred
        """
        session = self._dbSessionCreator()
        try:
            deviceInfo = (
                session.query(DeviceInfoTable)
                .filter(DeviceInfoTable.deviceId == action.deviceId)
                .one()
            )

            deviceId = deviceInfo.deviceId

            deviceInfo.lastOnline = action.dateTime
            deviceInfo.deviceStatus = action.deviceStatus

            session.commit()

            self._notifierController.notifyDeviceInfo(deviceId=deviceId)
            self._notifierController.notifyDeviceOnline(
                deviceInfo.deviceId, deviceInfo.deviceToken, deviceInfo.deviceStatus
            )

            return []

        except NoResultFound:
            return []

        finally:
            session.close()

    def _setAllDevicesOfflineBlocking(self):
        """Set All Devices to Offline"""
        session = self._dbSessionCreator()
        try:
            session.execute(
                DeviceInfoTable.__table__.update().values(
                    deviceStatus=DeviceInfoTuple.DEVICE_OFFLINE
                )
            )
            session.commit()

        finally:
            session.close()



from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.PluginNames import deviceTuplePrefix
from peek_core_device._private.storage.DeviceUpdateTuple import \
    DeviceUpdateTuple


@addTupleType
class CreateDeviceUpdateAction(TupleActionABC):
    """Create Update Action

    Create a new device update entry, this will be accompanied by an upload.

    """

    __tupleType__ = deviceTuplePrefix + "CreateDeviceUpdateAction"

    #:  Description of date1
    newUpdate: DeviceUpdateTuple = TupleField()



import logging
import shutil
import zipfile
from datetime import datetime
from pathlib import Path
from typing import List

import pytz
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from twisted.internet.defer import returnValue
from pytmpdir.spooled_named_temporary_file import SpooledNamedTemporaryFile
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Tuple import Tuple
from vortex.TupleAction import TupleActionABC

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.DeviceUpdateTuple import (
    DeviceUpdateTuple,
)
from peek_core_device._private.tuples.AlterDeviceUpdateAction import (
    AlterDeviceUpdateAction,
)
from peek_core_device._private.tuples.CreateDeviceUpdateAction import (
    CreateDeviceUpdateAction,
)
from peek_core_device._private.tuples.UpdateAppliedAction import (
    UpdateAppliedAction,
)

logger = logging.getLogger(__name__)


class UpdateController:
    def __init__(
        self,
        dbSessionCreator,
        notifierController: NotifierController,
        deviceUpdateFilePath: Path,
    ):
        self._dbSessionCreator = dbSessionCreator
        self._notifierController = notifierController
        self._deviceUpdateFilePath = deviceUpdateFilePath

    def shutdown(self):
        pass

    def processTupleAction(self, tupleAction: TupleActionABC) -> List[Tuple]:

        if isinstance(tupleAction, AlterDeviceUpdateAction):
            return self._processAdminAlter(tupleAction)

        if isinstance(tupleAction, UpdateAppliedAction):
            return self._processDeviceUpdated(tupleAction)

    @deferToThreadWrapWithLogger(logger)
    def _processAdminAlter(
        self, action: AlterDeviceUpdateAction
    ) -> List[Tuple]:
        """Process Admin Update

        :rtype: Deferred
        """
        session = self._dbSessionCreator()
        try:
            deviceUpdate = (
                session.query(DeviceUpdateTuple)
                .filter(DeviceUpdateTuple.id == action.updateId)
                .one()
            )

            deviceType = deviceUpdate.deviceType

            if action.remove:
                session.delete(deviceUpdate)
            else:
                deviceUpdate.isEnabled = action.isEnabled

            session.commit()

            self._notifierController.notifyDeviceUpdate(deviceType=deviceType)

            return []

        finally:
            # Always close the session after we create it
            session.close()

    @inlineCallbacks
    def processCreateUpdateUpload(
        self,
        namedTempFile: SpooledNamedTemporaryFile,
        action: CreateDeviceUpdateAction,
    ):
        """Process Create Update Upload

        Unlike the other action processes in the controllers, this method is called by
        a http resource upload.
        """

        if not zipfile.is_zipfile(namedTempFile.name):
            raise Exception("Uploaded archive is not a zip file")

        # Create the file name
        filePath = "%s/%s.zip" % (
            action.newUpdate.deviceType,
            action.newUpdate.updateVersion,
        )
        action.newUpdate.filePath = filePath
        action.newUpdate.urlPath = filePath.replace(r"\\", r"/")
        action.newUpdate.fileSize = Path(namedTempFile.name).stat().st_size

        # Create the database object, If that fails from some integrity problem
        # Then the file will delete it's self still
        deviceUpdateTuple = yield self._createUpdateOrmObj(action.newUpdate)

        absFilePath = self._deviceUpdateFilePath / filePath
        absFilePath.parent.mkdir(parents=True, exist_ok=True)

        shutil.move(namedTempFile.name, str(absFilePath))
        namedTempFile.delete = False

        self._notifierController.notifyDeviceUpdate(
            deviceType=deviceUpdateTuple.deviceType
        )

        returnValue(
            "%s:%s Created Successfully"
            % (deviceUpdateTuple.deviceType, deviceUpdateTuple.updateVersion)
        )

    @deferToThreadWrapWithLogger(logger)
    def _createUpdateOrmObj(
        self, newUpdate: DeviceUpdateTuple
    ) -> DeviceUpdateTuple:
        """Process Device Enrolment

        :rtype: Deferred
        """
        ormSession = self._dbSessionCreator()
        try:
            newUpdate.buildDate = datetime.now(pytz.utc)
            newUpdate.isEnabled = False
            ormSession.add(newUpdate)
            ormSession.commit()

            ormSession.refresh(newUpdate)
            ormSession.expunge_all()
            return newUpdate

        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def _processDeviceUpdated(self, action: UpdateAppliedAction) -> List[Tuple]:
        """Process Device Updated

        This action is sent when the device has applied an update, or attempted to.

        :rtype: Deferred
        """
        session = self._dbSessionCreator()
        try:
            deviceInfo = (
                session.query(DeviceInfoTable)
                .filter(DeviceInfoTable.deviceId == action.deviceId)
                .one()
            )

            deviceId = deviceInfo.deviceId

            if action.appVersion is not None:
                deviceInfo.appVersion = action.appVersion

            if action.updateVersion is not None:
                deviceInfo.updateVersion = action.updateVersion

            session.commit()

            self._notifierController.notifyDeviceInfo(deviceId=deviceId)

            return []

        finally:
            session.close()



import logging

from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.sql.schema import Index
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class DeviceUpdateTuple(Tuple, DeclarativeBase):
    """DeviceUpdateTuple

    This table stores information about the peek device updates.

    """

    __tablename__ = "DeviceUpdate"
    __tupleType__ = deviceTuplePrefix + "DeviceUpdateTuple"

    id = Column(Integer, primary_key=True, autoincrement=True)
    deviceType = Column(String(20), nullable=False)
    description = Column(String, nullable=False)
    buildDate = Column(DateTime(True), nullable=False)
    appVersion = Column(String(15), nullable=False)
    updateVersion = Column(String(15), nullable=False)
    filePath = Column(String(150), nullable=False)
    urlPath = Column(String(150), nullable=False)
    fileSize = Column(Integer, nullable=False)
    isEnabled = Column(Boolean, nullable=False, server_default="0")

    __table_args__ = (
        Index(
            "idx_DeviceUpdate_Version",
            deviceType,
            appVersion,
            updateVersion,
            unique=True,
        ),
    )



import logging

from txhttputil.site.FileUnderlayResource import FileUnderlayResource

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.server.update_resources.DeviceUpdateUploadResource import (
    DeviceUpdateUploadResource,
)
from peek_core_device._private.storage import DeclarativeBase
from peek_core_device._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_device._private.tuples import loadPrivateTuples
from peek_core_device.tuples import loadPublicTuples
from peek_plugin_base.server.PluginLogicEntryHookABC import (
    PluginLogicEntryHookABC,
)
from peek_plugin_base.server.PluginServerStorageEntryHookABC import (
    PluginServerStorageEntryHookABC,
)
from .DeviceApi import DeviceApi
from .TupleActionProcessor import makeTupleActionProcessorHandler
from .TupleDataObservable import makeTupleDataObservableHandler
from .admin_backend import makeAdminBackendHandlers
from .controller.MainController import MainController
from .controller.OfflineCacheController import OfflineCacheController

logger = logging.getLogger(__name__)


class LogicEntryHook(PluginLogicEntryHookABC, PluginServerStorageEntryHookABC):
    def __init__(self, *args, **kwargs):
        """ " Constructor"""
        # Call the base classes constructor
        PluginLogicEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

        self._api = None
        self._deviceUpdatesPath = None

    @property
    def dbMetadata(self):
        return DeclarativeBase.metadata

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        self._deviceUpdatesPath = (
            self.platform.fileStorageDirectory / "device_update"
        )
        self._deviceUpdatesPath.mkdir(parents=True, exist_ok=True)

        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()
        logger.debug("Loaded")

    def start(self):
        """Start

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        offlineCacheController = OfflineCacheController(
            dbSessionCreator=self.dbSessionCreator,
        )
        self._loadedObjects.append(offlineCacheController)

        tupleObservable = makeTupleDataObservableHandler(
            self.dbSessionCreator, offlineCacheController
        )
        self._loadedObjects.append(tupleObservable)

        notifierController = NotifierController(tupleObservable=tupleObservable)
        self._loadedObjects.append(notifierController)

        offlineCacheController.setNotificationController(notifierController)

        mainController = MainController(
            dbSessionCreator=self.dbSessionCreator,
            notifierController=notifierController,
            offlineCacheController=offlineCacheController,
            deviceUpdateFilePath=self._deviceUpdatesPath,
            tupleObservable=tupleObservable,
        )
        self._loadedObjects.append(mainController)

        # Support uploads from the admin UI
        # noinspection PyTypeChecker
        self.platform.addAdminResource(
            b"create_device_update",
            DeviceUpdateUploadResource(mainController.deviceUpdateController),
        )

        # Add the resource that the client uses to download the updates from the server
        updateDownloadResource = FileUnderlayResource()
        updateDownloadResource.addFileSystemRoot(str(self._deviceUpdatesPath))
        # noinspection PyTypeChecker
        self.platform.addServerResource(
            b"device_update", updateDownloadResource
        )

        self._loadedObjects.extend(
            makeAdminBackendHandlers(tupleObservable, self.dbSessionCreator)
        )

        # Make the Action Processor Handler
        self._loadedObjects.append(
            makeTupleActionProcessorHandler(mainController)
        )

        # Initialise the API object that will be shared with other plugins
        self._api = DeviceApi(mainController, self.dbSessionCreator)
        self._loadedObjects.append(self._api)

        notifierController.setApi(self._api)

        logger.debug("Started")

    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        self._api = None

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")

    @property
    def publishedServerApi(self) -> object:
        """Published Server API

        :return  class that implements the API that can be used by other Plugins on this
        platform service.
        """
        return self._api



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix
from peek_core_device._private.tuples.OfflineCacheStatusTuple import (
    OfflineCacheStatusTuple,
)


@addTupleType
class DeviceCacheStatusTuple(Tuple):
    """Device Cache Status Tuple

    This is an Admin tuple

    """

    __tupleType__ = deviceTuplePrefix + "DeviceCacheStatusTuple"

    statusList: list[OfflineCacheStatusTuple] = TupleField()



"""added gps tracking

Peek Plugin Database Migration Script

Revision ID: 02f9305f9e7f
Revises: 6bf6ecef3108
Create Date: 2021-02-22 15:17:11.886057

"""

# revision identifiers, used by Alembic.
revision = "02f9305f9e7f"
down_revision = "6bf6ecef3108"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "GpsLocation",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("deviceToken", sa.String(length=50), nullable=False),
        sa.Column("latitude", sa.Float(), nullable=False),
        sa.Column("longitude", sa.Float(), nullable=False),
        sa.Column("updatedDate", sa.DateTime(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["deviceToken"],
            ["core_device.DeviceInfo.deviceToken"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("deviceToken"),
        schema="core_device",
    )
    op.create_table(
        "GpsLocationHistory",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("deviceToken", sa.String(length=50), nullable=False),
        sa.Column("latitude", sa.Float(), nullable=False),
        sa.Column("longitude", sa.Float(), nullable=False),
        sa.Column("loggedDate", sa.DateTime(timezone=True), nullable=False),
        sa.ForeignKeyConstraint(
            ["deviceToken"],
            ["core_device.DeviceInfo.deviceToken"],
        ),
        sa.PrimaryKeyConstraint("id", "loggedDate"),
        schema="core_device",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("GpsLocationHistory", schema="core_device")
    op.drop_table("GpsLocation", schema="core_device")
    # ### end Alembic commands ###



from vortex.handler.TupleActionProcessorProxy import TupleActionProcessorProxy

from peek_core_device._private.PluginNames import deviceActionProcessorName
from peek_core_device._private.PluginNames import deviceFilt
from peek_plugin_base.PeekVortexUtil import peekServerName


def makeTupleActionProcessorProxy():
    return TupleActionProcessorProxy(
        tupleActionProcessorName=deviceActionProcessorName,
        proxyToVortexName=peekServerName,
        additionalFilt=deviceFilt,
    )



"""timezone aware

Peek Plugin Database Migration Script

Revision ID: 6bf6ecef3108
Revises: 5dd43ff7ea8e
Create Date: 2017-12-31 14:28:47.653836

"""

# revision identifiers, used by Alembic.
revision = "6bf6ecef3108"
down_revision = "5dd43ff7ea8e"
branch_labels = None
depends_on = None

from alembic import op


def upgrade():
    # core_device."DeviceInfo"
    op.execute(
        """ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "lastOnline" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastOnline" AT TIME ZONE 'UTC'
               """
    )

    op.execute(
        """ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "lastUpdateCheck" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastUpdateCheck" AT TIME ZONE 'UTC'
               """
    )

    op.execute(
        """ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "createdDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "createdDate" AT TIME ZONE 'UTC'
               """
    )

    # core_device."DeviceUpdate"
    op.execute(
        """ALTER TABLE core_device."DeviceUpdate"
                  ALTER COLUMN "buildDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "buildDate" AT TIME ZONE 'UTC'
               """
    )


def downgrade():
    pass



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix


@addTupleType
class OfflineCacheSettingTuple(Tuple):
    """Offline Cache Setting Tuple

    This tuple is for the client UI settings.

    """

    __tupleType__ = deviceTuplePrefix + "OfflineCacheSettingTuple"

    offlineEnabled: bool = TupleField(defaultValue=False)
    offlineCacheSyncSeconds: int = TupleField(defaultValue=15 * 60)



import logging

from twisted.internet.defer import inlineCallbacks

from peek_core_device._private.client.DeviceOnlineHandler import (
    DeviceOnlineHandler,
)
from peek_core_device._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_device._private.tuples import loadPrivateTuples
from peek_core_device.tuples import loadPublicTuples
from peek_plugin_base.client.PluginClientEntryHookABC import (
    PluginClientEntryHookABC,
)
from .DeviceTupleDataObservableProxy import makeDeviceTupleDataObservableProxy
from .DeviceTupleProcessorActionProxy import makeTupleActionProcessorProxy

logger = logging.getLogger(__name__)


class ClientEntryHook(PluginClientEntryHookABC):
    def __init__(self, *args, **kwargs):
        """ " Constructor"""
        # Call the base classes constructor
        PluginClientEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

    def load(self) -> None:
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        loadStorageTuples()

        loadPrivateTuples()
        loadPublicTuples()

        logger.debug("Loaded")

    def start(self):
        """Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """

        # ----------------
        # Setup Photo Resource Proxy
        # Support file downloads for device updates
        # noinspection PyTypeChecker
        proxyResource = self.createProxy()
        # Matches resource path on server
        # noinspection PyTypeChecker
        self.platform.addOfficeResource(b"device_update", proxyResource)

        # noinspection PyTypeChecker
        self.platform.addFieldResource(b"device_update", proxyResource)

        # ----------------
        # Action Processor Proxy
        self._loadedObjects.append(makeTupleActionProcessorProxy())

        # ----------------
        # Data Observable Proxy
        self._loadedObjects.append(makeDeviceTupleDataObservableProxy())

        # ----------------
        # Online Handler
        self._loadedObjects.append(DeviceOnlineHandler())

        # # ----------------
        # # Update Download Handler
        # self._loadedObjects.append(
        #     UpdateDownloadHandler(self.platform.fileStorageDirectory,
        #                           self.platform.peekServerHost,
        #                           self.platform.peekServerHttpPort)
        # )

        logger.debug("Started")

    @inlineCallbacks
    def stop(self):
        """Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            yield self._loadedObjects.pop().shutdown()

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")



import logging

from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Float
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.sql.schema import ForeignKey
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix
from .DeclarativeBase import DeclarativeBase
from ...tuples.DeviceGpsLocationTuple import DeviceGpsLocationTuple

logger = logging.getLogger(__name__)


@addTupleType
class GpsLocationTable(Tuple, DeclarativeBase):
    __tablename__ = "GpsLocation"
    __tupleType__ = deviceTuplePrefix + "GpsLocationTable"

    id = Column(Integer, primary_key=True)
    deviceToken = Column(
        String(50),
        ForeignKey("DeviceInfo.deviceToken", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )
    latitude = Column(Float, nullable=False)
    longitude = Column(Float, nullable=False)
    updatedDate = Column(DateTime(True), nullable=False)

    def toTuple(self):
        return DeviceGpsLocationTuple(
            deviceToken=self.deviceToken,
            latitude=self.latitude,
            longitude=self.longitude,
            datetime=self.updatedDate,
        )



import logging
from collections import namedtuple

from peek_core_device._private.server.controller.NotifierController import (
    NotifierController,
)
from peek_core_device._private.storage.GpsLocationHistoryTable import (
    GpsLocationHistoryTable,
)
from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable
from peek_core_device._private.tuples.UpdateDeviceGpsLocationTupleAction import (
    UpdateDeviceGpsLocationTupleAction,
)
from sqlalchemy.dialects.postgresql import insert
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.TupleAction import TupleActionABC
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

logger = logging.getLogger(__name__)
DeviceLocationTuple = namedtuple(
    "DeviceLocationTuple",
    ["deviceToken", "latitude", "longitude", "updatedDate"],
)
TimezoneSetting = namedtuple("TimezoneSetting", ["timezone"])


class GpsController(TupleActionProcessorDelegateABC):
    def __init__(
        self,
        dbSessionCreator,
        tupleObservable: TupleDataObservableHandler,
        notifierController: NotifierController,
    ):
        self._dbSessionCreator = dbSessionCreator
        self._tupleObservable = tupleObservable
        self._notifierController = notifierController

    def shutdown(self):
        pass

    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        if isinstance(tupleAction, UpdateDeviceGpsLocationTupleAction):
            return self._processGpsLocationUpdateTupleAction(tupleAction)

    @inlineCallbacks
    def _processGpsLocationUpdateTupleAction(
        self, action: UpdateDeviceGpsLocationTupleAction
    ):
        yield
        # capturedDate = datetime.now()
        currentLocation = DeviceLocationTuple(
            deviceToken=action.deviceToken,
            latitude=action.latitude,
            longitude=action.longitude,
            updatedDate=action.datetime,
        )
        updatedGpsLocationTableRow = self._updateCurrentLocation(
            currentLocation
        )
        self._logToHistory(currentLocation)
        self._notifyTuple(updatedGpsLocationTableRow)
        return []

    def _notifyTuple(self, gpsLocationTable: GpsLocationTable):
        tuple_ = gpsLocationTable.toTuple()
        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(
                tuple_.tupleName(),
                dict(deviceToken=tuple_.deviceToken),
            )
        )

        self._notifierController.notifyDeviceGpsLocation(
            gpsLocationTable.deviceToken,
            gpsLocationTable.latitude,
            gpsLocationTable.longitude,
            updatedDate=gpsLocationTable.updatedDate,
        )

    def _updateCurrentLocation(
        self, currentLocation: DeviceLocationTuple
    ) -> GpsLocationTable:
        statement = insert(GpsLocationTable).values(currentLocation._asdict())
        statement = statement.on_conflict_do_update(
            index_elements=[GpsLocationTable.deviceToken],
            set_=currentLocation._asdict(),
        )
        session = self._dbSessionCreator()
        try:
            # upsert updates
            r = session.execute(statement)
            insertedPrimaryKey = r.inserted_primary_key[0]
            session.commit()

            if not insertedPrimaryKey:
                return

            # get the affected row
            updatedRows = session.query(GpsLocationTable).filter(
                GpsLocationTable.id == r.inserted_primary_key[0]
            )
            result = updatedRows.one()
            session.expunge_all()
            return result
        finally:
            session.rollback()
            session.close()

    def _logToHistory(self, currentLocation: DeviceLocationTuple):
        session = self._dbSessionCreator()
        record = GpsLocationHistoryTable(
            deviceToken=currentLocation.deviceToken,
            latitude=currentLocation.latitude,
            longitude=currentLocation.longitude,
            loggedDate=currentLocation.updatedDate,
        )
        try:
            session.add(record)
            session.commit()
        finally:
            session.close()



import logging
from typing import Union

from peek_core_device._private.server.controller.OfflineCacheController import (
    OfflineCacheController,
)
from peek_core_device._private.storage.DeviceInfoTable import DeviceInfoTable
from peek_core_device._private.storage.GpsLocationTable import GpsLocationTable
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class DeviceInfoTupleProvider(TuplesProviderABC):
    def __init__(
        self, ormSessionCreator, offlineCacheController: OfflineCacheController
    ):
        self._ormSessionCreator = ormSessionCreator
        self._offlineCacheController = offlineCacheController

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:

        deviceId = tupleSelector.selector.get("deviceId")

        ormSession = self._ormSessionCreator()
        try:
            query = ormSession.query(
                DeviceInfoTable, GpsLocationTable
            ).outerjoin(GpsLocationTable)

            if deviceId is not None:
                query = query.filter(DeviceInfoTable.deviceId == deviceId)

            tuples = []
            for deviceInfoTableRow, gpsLocationTableRow in query.all():
                tuples.append(
                    deviceInfoTableRow.toTuple(
                        currentLocationTuple=gpsLocationTableRow,
                        lastCacheCheck=self._offlineCacheController.lastCacheUpdate(
                            deviceInfoTableRow.deviceToken
                        ),
                    )
                )

            # Create the vortex message
            return (
                Payload(filt, tuples=tuples).makePayloadEnvelope().toVortexMsg()
            )

        finally:
            ormSession.close()



from abc import ABCMeta
from abc import abstractmethod
from typing import List
from typing import Optional

from rx import Observable
from twisted.internet.defer import Deferred

from peek_core_device.tuples.DeviceGpsLocationTuple import DeviceGpsLocationTuple


class DeviceApiABC(metaclass=ABCMeta):
    """Device API

    This is the public API for the part of the plugin that runs on the server service.

    """

    @abstractmethod
    def deviceDetails(self, deviceTokens: List[str]) -> Deferred:
        """Device Details

        Retrieve the details for the devices with the device tokens provided.

        :param deviceTokens: A list of device token strings
        :return: A Deferred that will fire with :code:`List[DeviceDetailTuple]`

        """

    @abstractmethod
    def deviceDescription(self, deviceToken: str) -> Deferred:
        """Device Description

        Retrieve the devices description

        :param deviceToken: The token for the device to retrieve the description for
        :return: A Deferred that will fire with :code:`Optional[str]`

        """

    @abstractmethod
    def deviceDescriptionBlocking(self, deviceToken: str) -> Optional[str]:
        """Device Description

        Retrieve the devices description, this must be called from a thread.

        :param deviceToken: The token for the device to retrieve the description for
        :return: The device description or None

        """

    @abstractmethod
    def deviceOnlineStatus(self) -> Observable:
        """Device Online Stauts

        Subscribe to device online status

        :return: An observable that fires when devices go online or offline.

        """

    @abstractmethod
    def deviceCurrentGpsLocation(self) -> Observable:
        """Device Current GPS Location

        Subscribe to device current GPS location
        :return: An observable that fires when devices update GPS locations
        """

    @abstractmethod
    def deviceCurrentGpsLocations(
        self, deviceTokens: List[str]
    ) -> List[DeviceGpsLocationTuple]:
        """Device Current GPS Locations from a list device tokens

        :param deviceTokens:  A list of string
        :return: A list of DeviceGpsLocationTuple
        """

    @abstractmethod
    def deviceTokens(self) -> List[str]:
        """List all device tokens in str

        :return: A list of all device tokens in str
        """
